from .doc2vec import GensimDoc2Vec
from .fasttext import GensimFastText
from .latent_semantic_analysis import GensimLatentSemanticAnalysis
from .random_indexing import GensimRandomIndexing
from .word2vec import GensimWord2Vec
from .lda import GensimLDA
